/****************************************************************************
*
*                   Digital Sound Interface Kit (DSIK)
*                            Version 2.00
*
*                           by Carlos Hasan
*
* Filename:     example5.c
* Version:      Revision 1.0
*
* Language:     WATCOM C
* Environment:  IBM PC (DOS/4GW)
*
* Description:  Speed benchmark example program.
*               The number of ticks per frame will change drastically
*               based on the music module used, some modules are much
*               more CPU intensive than others.
*
* IMPORTANT:    This program should be executed from the DOS prompt,
*               do not try to execute it under multitasking systems
*               (ie. Windows, OS/2). Also, remove all the TSRs from
*               memory, specially disk cache programs.
*
****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <conio.h>
#include "audio.h"
#include "import.h"
#include "timer.h"

/* 16-bit stereo 44.1khz 16 channels = 16% of CPU time (486DX2/66) */

#define NUMCHANS 4
#define MAXFRAMES 1000

#define MODPATH "SONG.DSM"
#define MODFORM FORM_DSM

#define SMPPATH "SAMPLE.WAV"
#define SMPFORM FORM_WAV

int frames,busy,ticks[MAXFRAMES];

void setborder(int color)
{
    inp(0x3da);
    outp(0x3c0,0x31);
    outp(0x3c0,color);
}

void starttimer(void)
{
    outp(0x61,inp(0x61) | 0x01);
    outp(0x43,0xb4);
    outp(0x42,0xff);
    outp(0x42,0xff);
}

int readtimer(void)
{
    int n;
    outp(0x43,0xb0);
    n = inp(0x42);
    n += inp(0x42) << 8;
    return 0x10000L - n;
}

void timerhandler(void)
{
    if (!busy) {
        busy = 1;
        setborder(1);
        starttimer();
        dPoll();
        if (frames < MAXFRAMES) ticks[frames++] = readtimer();
        setborder(0);
        busy = 0;
    }
}

void benchmark(SoundCard *SC, int NumChans)
{
    int i;
    double u,v;

    cprintf("Please wait... ");
    busy = frames = 0;
    while (!kbhit() && (frames < MAXFRAMES))
        cprintf("(%02d%%)\b\b\b\b\b", (100*frames)/MAXFRAMES);
    if (kbhit()) getch();
    cprintf("\r");

    /* compute the amount of time required per frame */
    u = 0.0;
    for (i = 0; i < frames; i++) u += (double)ticks[i];
    u = (u/frames);
    v = 0.0;
    for (i = 0; i < frames; i++) v += (double)(ticks[i]-u)*(ticks[i]-u);
    v = sqrt(v/frames);
    i = ((SC->Modes & AF_16BITS ? 1 : 0) +
            (SC->Modes & AF_STEREO ? 1 : 0));
    printf( "Number of channels  : %d\n"
            "Sample rate         : %d\n"
            "Bytes per second    : %d\n"
            "Ticks per frame     : %4.2f \361 %4.2f\n"
            "CPU time required   : %2.2f \361 %4.2f %%\n\n",
            NumChans, SC->SampleRate, (long)SC->SampleRate << i,
            u,v, (100.0*u)/TICKS(70), (100.0*v)/TICKS(70));
}

int main(void)
{
    SoundCard SC;
    Sample *S;
    DSM *M;
    int i;
    float u,v;

    dRegisterDrivers();
    if (dLoadSetup(&SC,"SETUP.CFG")) {
        printf("Please run SETUP.EXE to configure.\n");
        exit(EXIT_FAILURE);
    }
    if (SC.ID == ID_GUS) {
        printf("Sorry, this program does not work for GUS soundcards.\n");
        exit(EXIT_FAILURE);
    }
    if (dInit(&SC)) {
        printf("Error initializing the sound system.\n");
        exit(EXIT_FAILURE);
    }
    atexit((void(*)(void))dDone);
    dInitTimer();
    atexit((void(*)(void))dDoneTimer);
    dStartTimer(timerhandler,0);

    /* (TEST1) load sample file and enable the looping mode */
    printf("Playing sample file : %s\n",SMPPATH);
    if (!(S = dImportSample(SMPPATH,SMPFORM))) {
        printf("Error (%03d) loading %s sample file: %s.\n",
            dError, SMPPATH, dErrorMsg[dError]);
        exit(EXIT_FAILURE);
    }
    S->Flags |= SF_LOOPED;
    S->LoopStart = 0; S->LoopEnd = S->Length;

    /* start playing the sample on NUMCHANS channels simultaneously */
    dSetupVoices(NUMCHANS,128);
    for (i = 0; i < NUMCHANS; i++) {
        dPlayVoice(i,S);
        dSetVoiceBalance(i, i & 1 ? PAN_RIGHT : PAN_LEFT);
    }

    /* sample time slices required by the sound system */
    benchmark(&SC,NUMCHANS);

    /* stop all the voices and release the sample data */
    for (i = 0; i < NUMCHANS; i++) dStopVoice(i);
    dFreeSample(S);


    /* (TEST2) load module file and enable the looping */
    printf("Playing module file : %s\n",MODPATH);
    if (!(M = dImportModule(MODPATH,MODFORM))) {
        printf("Error (%03d) loading %s module file: %s.\n",
            dError, MODPATH, dErrorMsg[dError]);
        exit(EXIT_FAILURE);
    }
    M->Header.ReStart = 0;

    /* start playing the sample on NUMCHANS channels simultaneously */
    dSetupVoices(M->Header.NumTracks,M->Header.MasterVolume);
    dPlayMusic(M);

    /* sample time slices required by the sound system */
    benchmark(&SC,M->Header.NumTracks);

    /* stop music and release the module */
    dStopMusic();
    dFreeModule(M);

    dStopTimer();
    return 0;
}
